<?php
/*
# Google Map API for PmWiki 
# =========================
# 
# Copyright Statement 
# -------------------
# 
# Copyright (c) 2006, Benjamin C. Wilson. All Rights Reserved. You have
# permission to use this software for web site use provided this copyright
# statement remains intact. You may not otherwise republish or alter this
# software without prior permission of the author. License to publish granted
# PmWiki.org for distribution to the PmWiki community. This copyright statement
# must accompany the script.  Any exercise of rights under this license by you
# is subject to the following conditions.
#    1. Redistribution of this software is not permitted.
#    2. Any user documentation must include the copyright statement.
#    3. This copyright statement, disclaimers and limitations must remain with
#       the software.
#    4. Consent to the following disclaimer and limitations of liability.
# 
# You may contact the author at ameen@dausha.net. This software is maintained at
# http://www.dausha.net/GoogleMapAPI/. If you have any recommended changes,
# please contact the author.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
# FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDER AND
# CONTRIBUTORS MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
# ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# Limitations of Liability
# ~~~~~~~~~~~~~~~~~~~~~~~~
#
# THE COPYRIGHT HOLDER AND CONTRIBUTORS SHALL HAVE NO LIABILITY TO LICENSEE OR
# OTHER PERSONS FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, CONSEQUENTIAL,
# EXEMPLARY, OR PUNITIVE DAMAGES OF ANY CHARACTER INCLUDING, WITHOUT
# LIMITATION, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR
# PROFITS, OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY OF
# CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT LIABILITY OR
# OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
#
# License Revisions
# ~~~~~~~~~~~~~~~~~
#
# The Software Copyright Holder may publish revised and/or new versions of this
# this License from time to time. Each version will be given a distinguishing
# version number. Once a version of Software has been published under a
# particular version of this License, you may continue to use it under the
# terms of that license version. You may also choose to use such Software under
# the terms of any subsequent version of this License published by the
# Copyright Holder. No one other than the Copyright Holder has the right to
# modify the terms of this License.
# 
# Abstract
# --------
# Embedding a Google map is one way for a web site to speak to its visitors.
# However, generating a Google map is a complex ordeal made slightly more
# difficult when using a wiki to manage site content. The Google Map API (GMA)
# is a receipe for PmWiki that helps the site editor seamlessly make and
# maintain a Google map on the site. This is accomplished by using an
# integrated suite of PmWiki directives which allows editors to mark points,
# make lines, and tie a link to a point on the map. The result is a simple way
# to use Google's powerful map protocol on a PmWiki site.
# 
# Installation
# ------------
#
# Install the Script. The first step in the installation is to acquire and
# install the software. To place the files in their proper location, extract
# the files in the parent directory to both the /pub and /cookbook
# directories.  Otherwise, copy the contents of the two directories as
# appropriate.  Specifically, this software should have the recipe file in the
# /cookbook directory, the javascript in the /pub/scripts directory, and images
# in /pub/images/gma directory.
#
# Configure the Site. After the code is installed, you will need to configure
# the site. This is typically done in the local/config.php file, so this How-To
# assumes an installation using the file. Before the site is configured,
# however, you will need to contact Google to obtain a site key.
# (http://www.google.com/apis/maps/signup.html. You will need a unique key for
# each distinct web site. Google does not provide its interface to sites
# without this key.
#
# In the configuration file (local/config.php), you must set one variable:
# $GmaKey. Beyond that, you may set other values using $GmaDefaults as
# discussed below. These values should be set before the recipes called. Most
# of these defaults may be overridded via the directives.
#
# Once the variables are set, you will need to include the recipe (e.g.
# "require_once('path/to/cookbook/gma-2.0.0.php);"). This command must follow
# the $GmaKey and $GmaDefaults variable declarations.
#
# Configure the Skin. Two variables must be embedded in the PmWiki skins that
# are used on that site: $GmaScript and $GmaIEFix. The former variable,
# $GmaScript, must be added immediate before the </body> tag (e.g.
# $GmaScript</body></html>)---in order to avoid triggering a browser bug.
# $GmaIEFix must be placed within the leading <html> tag (e.g.
# <html$GmaIEFix>) to offset a separate browser bug. (Oddly enough, both bugs
# are Internet Explorer-based.) Additionally, you may want to set a CSS value
# for the default DIV tag (e.g. <div id='map'>) in that skin. This is not
# required as the recipe sets default values, but it may be desired.
# 
# Using the Google Map API for PmWiki
# -----------------------------------
#
# Using this recipe is fairly straight forward. There are three types of map
# tools available: point, line, and map. The map tool must be used to generate
# the map. Either the point or the line must also be used, and these types may
# be mixed. We will take each map tool in turn, beginning with the map tool.
#
# Using the Map Tool. The map tool generates the map itself. The map is located
# where the directive is placed on the wiki page. Specifically, the map tool
# returns an empty DIV statement which the Google software uses to embed the
# map. The map tool provides several options which help control the use of the
# map. Each of these options may be configured by the site administrator, or
# set by the editor for a specific map. These options include:
#    * 'id' (default map): The id option allows the editor to determine the CSS
#      name for the map. When the multiple-maps feature is implemented, this
#      will allow multiple maps to be generated.
#    * 'view' (default hybrid): The view obtion allows the editor to set the
#      initial display of the map. There are three displays available: normal,
#      or a traditional map view; satellite, or a map showing only a satellite
#      image mosaic; and hybrid, or a map which overlays political and road
#      lines over a satellite map. These displays are represented by
#      'nor','sat', or 'hyb' respectively.
#    * 'height' (default 300px): The height option the editor to set the height
#      of the map. This value responds to CSS values (e.g. pixels, em, pt,
#      percent, etc.)
#    * 'width' (default 500px): The width option the editor to set the width of
#      the map. This value responds to CSS values (e.g. pixels, em, pt,
#      percent, etc.)
#    * 'ctrl' (default ): The ctrl option allows the editor to select which map
#      controls are available to the visitor. There are () controls: maptype,
#      large, small, and overview. As each map control may be set by the site
#      administrator, the editor may unset a default value. The negative sign
#      is used to unset a default value (e.g. if the default is to show the
#      overview, then '-overview' will turn off this map control. This variable
#      is an array and values are set in serial (e.g.
#      "ctrl=small,-overview,sat")
#      * The maptype control allows the visitor to change the map type between
#        hybrid, normal and satellite. If set by default, -maptype will turn
#        this control off.
#      * The large and small controls allow the editor to decide whether the
#        zoom and pan controls are large or small. Setting the opposite will
#        deselect the default size.
#      * The overview control allows the visitor to see a small window in the
#        lower right hand corner. This window displays a smaller-scale map of
#        the area in the main window.

# Using the Point Tool. The point tool places a marker at the designated spot.
# The location is determined by latitude and longitude settings. Additionally,
# the point can contain a link and/or text. (Coming soon) Finally, the marker
# look can be determined using the icon option. The options include:
#    * lat (default 0): The lat option sets the point's latitude.
#    * lon (default 0): The lon option sets the point's latitude.
#    * zip (default 0): The zip option optionally determines the point's
#      latitude and longitude from the U.S. Census data file (circa 1999).
#    * link (no default): This option creates a link between the point on the
#      Google map and the link-text on the page. Clicking the link on the page
#      causes this point's information window to open. Clicking the link in the
#      information window sends the visitor to the link on the page. The link
#      text should be quoted.
#    * text (no default): This option creates HTML text in an information
#      window. Clicking this point's marker causes the information window to
#      open. The text should be quoted.
#    * icon (not active): This option allows the editor to use any of a
#      number of icons to distinguish points on the map. This includes letters,
#      numbers, colors and shapes.
#
# Using the Line Tool. The line tool generates color-coded lines drawn by a
# series of line-points. Each line-point along the line is described using this
# markup.  The location of each is determined by latitude and longitude
# settings. Additionally, the point can contain a link and/or text.
#    * id (default 0): The id option allows the editor to create multiple
#      lines. When creating more than one line, the id number designates to
#      which line the point is appended.
#    * lat (default 0): The lat option sets the line-point's latitude.
#    * lon (default 0): The lon option sets the point's latitude.
#    * zip (default 0): The zip option optionally determines the point's
#      latitude and longitude from the U.S. Census data file (circa 1999).
#    * 'color' (default #880088): The color option allows the editor to set
#      the line's color. Once the color is set for a given line, it need not be
#      designated again.
#    * 'opacity' (default 2): The opacity option allows the editor to set the
#      line's opacity, measured in percentage (e.g. 0 - 100 percent). Once the
#      opacity is set for a given line, it need not be designated again.
#    * 'thickness' (default 2): The thickness option allows the editor to set
#      the line thickness, measured in pixels. Once the thickness is set for a
#      given line, it need not be designated again.

*/
define('GMAPATH', dirname(__FILE__) . '/');
include_once(GMAPATH.'library/php-local-browscap.php');

# SDV()
SDV($GmaDebug, 0);
SDV($GmaDefaults, array(
    'ctrl' => array('large','overview'),
    'view' => 'hybrid',
    'height' => '300px',
    'width' => '500px',
    'float' => '',
    'color' => '#808',
    'thickness' => 2,
    'opacity' => 0.7,
    'zoom' => 0
    )
);
SDV($GmaKey, '');
# Internals.
$GmaDebugMsg = array();
$GmaHTML = '';
$GmaIEFix = '';
$GmaJS = '';
$GmapJS = '';
$GmaScript = '';
$GmaLines = array();
$GmaPoints = array();
$GmaPointPopulation = 0;
$GmaVersion = '2.0.0';
# Markup
# * (:gma-map [options]:)
# * (:gma-point lat lon [options]:)
# * (:gma-line lat lon [options]:)
Markup('gma', '_begin', '/\(:gma-(\S+)\s?(.*?):\)/ie',"gmaMarkup('$1','$2');");
class GmaLine
{
  function GmaLine($id) {
    global $GmaLines, $GmaDefaults;
    $this->id = $id;
    $this->color = $GmaDefaults['color'];
    $this->opacity = $GmaDefaults['opacity'];
    $this->points = array();
    $this->thickness = $GmaDefaults['thickness'];
  }
  function addPoint($lat, $lon) {
    if ($lat && $lon) array_push($this->points, "new GLatLng($lat, $lon)");
  }
  function line($map_id='map') {
    $id = $this->id;
    $pad = str_pad('', 26 + strlen($id), ' ', STR_PAD_LEFT);
    $points = '['.join(",\n$pad", $this->points).']';
    $opacity = ($this->opacity > 1) 
        ? sprintf("%0.2f", $this->opacity / 100)
        : $this->opacity;
    $opacity = ($opacity > 1) ? 1.0 : $opacity;
    $options = join(', ', array(
                    $points,
                    $this->color(),
                    $this->thickness,
                    $opacity
                ));
    $this->overlay = "    $map_id.addOverlay($id);\n";
    return "    var $id = new GPolyline($options);\n";
  }
  function color() {
    // This method returns only web-safe colors.
    $color = preg_replace("/\\\'/",'',$this->color);
    if (0 === strpos($color, '#')) $color = substr($color, 1);
    // break into hex 3-tuple
    $cutpoint = ceil(strlen($color) / 2) - 1;
    $rgb = explode(':', wordwrap($color, $cutpoint, ':', $cutpoint), 3);

    $out = '';
    foreach($rgb as $r) {
        if (strlen($r) == 1) $r .= $r; # Expand single value colors.
        $r = (isset($r)) ? hexdec($r) : 0; # Hex to Dec.
        $r = (round($r/51) * 51); # Make web-safe
        $out .= str_pad(dechex($r), 2, '0', STR_PAD_LEFT); # Code color.
    }
    return "'#$out'";
  }
}

class GmaMap
{
    var $map_views = array(
      'hyb' => 'G_HYBRID_MAP',
      'nor' => 'G_NORMAL_MAP',
      'sat' => 'G_SATELLITE_MAP',
    );
    var $controls_js = array(
      'large' => 'new GLargeMapControl()',
      'maptype' => 'new GMapTypeControl()', # allow toggle b/w map types.
      'overview' => 'new GOverviewMapControl()', # allow map controls.
      'small' => 'new GSmallMapControl()',
    );
    var $controls = array();
    function GmaMap($id='map') {
      global $GmaDefaults;
      $this->id = $id;
      $this->view = $GmaDefaults['view'];
      $this->height = $GmaDefaults['height'];
      $this->width = $GmaDefaults['width'];
      $this->zoom = 0;
    }
    function setControl($controls) {
      foreach ((array) $controls as $c) {
        if ($c[0] == '-' && array_key_exists(substr($c,1), $this->controls)) {
          unset($this->controls[substr($c,1)]);
        }
        else {
          $this->controls[$c] = 1;
        }
      }
    }
    function getControls() {
        $ret = '';
        $id = $this->id;
        foreach (array_keys($this->controls) as $c) {
          if ($c == 'small' && array_key_exists('large', $this->controls)){
            unset($this->controls[$c]);
          }
          elseif ($c == 'large' && array_key_exists('small', $this->controls)){
            unset($this->controls[$c]);
          }
          elseif ($c[0] != '-') {
            $c = $this->controls_js[$c];
            $ret .= "    $id.addControl($c);\n";
          }
        }
        return $ret;
    }
    function view() {
        return $this->map_views[strtolower(substr($this->view,0,3))];
    }
}
class GmaPoint
{
  var $directions = '';
  var $icon = '';
  var $lat = 0.0;
  var $lon = 0.0;
  var $message = '';
  var $title = '';
  function GmaPoint($lat='', $lon='') {
    global $GmaPointPopulation;
    $this->id = $GmaPointPopulation++;
    if ($lat && $lon) { $this->lat = $lat; $this->lon = $lon; }
  }
  function icon() { return $this->_quote($this->icon);}
  function point($map_id='map') {
    $icon = $this->icon();
    $lat = $this->lat;
    $lon = $this->lon;
    $text = $this->message();
    $title = $this->_quote($this->title);
    return "    addGmaPoint($map_id,$lat,$lon,$title,$text,$icon);\n";
  }
  function _quote($i) { return ($i) ? $this->_jsstrip("'$i'") : 'null'; }
  function _jsstrip($i) { return $i;} # TODO: Strip Javascript.
  function message() {
    if ($this->message) {
      $m = $this->message;
      $m = preg_replace('/&lt;/','<', $m);
      $m = preg_replace('/&gt;/','>', $m);
    }
    $m = $this->_quote($m);
    return $m;
  }
  function anchor($anchor,$mapid='map') {
    if (!$anchor) { return ''; }
    $anchor = $this->_jsstrip($anchor);
    $id = $this->id;
    $link =  "<a name='{$anchor}'></a>"
            ."<a href='javascript:makeGmalink({$mapid},{$id});'>{$anchor}</a>";
    $this->message .= "<a href=\'#$anchor\'>$anchor</a>";
    return "$link";
  }
} // End Class GmaPoint
function gmaBrowserFix() {
}
function gmaCleanup() {
  global $HTMLHeaderFmt, $HTMLStylesFmt;
  global $GmaEnable, $GmaVersion, $GmaScript;
  global $GmaMap, $GmaPoints, $GmaLines;
  global $GmaKey, $GmaDebugMsg, $GmaDefaults;

  GmaDoIEFix();
  #--------------------------------
  # Set the HTML
  $height = $GmaMap->height;
  $width = $GmaMap->width;
  $float = ($GmaMap->float) ? "float: {$GmaMap->float};" : '';
  $map_id = $GmaMap->id;

  $HTMLStylesFmt['gmap_api'] 
     = "div#$map_id{ height: $height; width:$width;$float}";
  $HTMLHeaderFmt[] 
     =  '<style type=\'text/css\'>v:* { behavior:url(#default#VML); }</style>'
       ."\n<script src='http://maps.google.com/maps?file=api&v=2&key="
       .$GmaKey."' type='text/javascript'></script>";
  $HTMLHeaderFmt[] 
     =  "\n<script language='javascript' src='\$PubDirUrl/scripts/gmaJs.js'>"
       ."</script>\n"; 

  GmaDebug(print_r($GmaMap, 1));
  $controls = $GmaMap->getControls();
  $view = $GmaMap->view();
  $zoom = $GmaMap->zoom or 'null';
  $overlay = '';
  $points = '';
  $lines = '';
  foreach ($GmaPoints as $p) { $points .= $p->point($map_id); }
  foreach ($GmaLines as $l) { 
      $lines .= $l->line($map_id); 
      $overlay .= $l->overlay;
  }
  $debug = ($GmaDebug) ? implode("\n", (array) $GmaDebugMsg) : '';
  $defaults = $GmaDefaults; 
  $defaults['ctrl'] = implode(', ',$defaults['ctrl']);
  $defaults = preg_replace('/Array|\(\\n|\)/', '', print_r($defaults, 1));

  $GmaScript =<<<GMASCRIPT
  $debug
  <!-- GMA Site Default Controls:
  $defaults -->
  <script type="text/javascript">
  //<![CDATA[
  // Copyright (c) 2006, Benjamin C. Wilson. All Rights Reserved.
  // Google Map API for PmWiki, $GmaVersion.
  // This copyright statement must accompany this script.
  if (GBrowserIsCompatible()) {
    var from_htmls = [];
    var htmls = [];
    var markers = [];
    var points = [];
    var to_htmls = [];
    var i = 0;
    var $map_id = new GMap2(document.getElementById('$map_id'));
    var bounds = new GLatLngBounds;
$points
$lines
    setGmaMapCenter($map_id, $view, $zoom);
    $map_id.addControl(new GScaleControl());
$controls
$overlay
    doGmaOverlay($map_id);
  }
</script>
GMASCRIPT;
}
function gmaDebug($m) {
  // (null) gmaDebug(string);
  //
  // Packs end-of-cycle debugging information.
  global $GmaDebug, $GmaDebugMsg;
  array_push($GmaDebugMsg, "<pre>$m</pre>");
}
function gmaDefaults($opt, $def, $orig='') {
  // (mixed) gmaDefaults(option, site-default);
  //
  // This function overrides site default variable.
  if (!($opt || $def || $org)) return '';
  $val = ($opt && !$orig) ? $opt : $def;
  $val = ($orig) ? $orig : ($opt) ? $opt : $def;
  return $val;
}
function gmaLLZError($type, $lat, $lon) {
  // (bool) gmaLLZError(type, lat, lon);
  //
  // function flags error when the type requires a lattitude or longitude.
  if ($type == 'map') return 0;
  if (!$lat && !$lon) return 1;
}
function gmaMarkup($type, $args) {
  // (string) gmaMarkup(type, args);
  //
  // This function converts all PmWiki markup into GMA objects and attributes
  // and reconciles changes to the site-default.

  global $GmaDefaults, $GmaDebug;
  gmaSpoofLocale();
  $ret = '';

  $opts = parseArgs($args);
  GmaDebug("OPTIONS:".print_r($opts,1));
  GmaDebug("CONTROLS:".print_r($controls,1));

  // GMA accepts either ZIP code or lat/lon.
  if ($opts['zip'])
    list($opts['lat'], $opts['lon']) = gmaZIP2LatLon($opts['zip']);
  if (gmaLLZError($type, $opts['lat'], $opts['lon']))
    return 'GMA Error: No Lat/Lon or ZIP given (or incomplete).';

  // This switch handles the behavior of the various GMA Types.
  switch($type) {
    // Gma Type: Line "(:gma-line lat lon [options]:)"
    case ('line'):
        global $GmaLines;
        $lid = GmaDefaults('line'.$opts['id'], 'line0');
        if (!array_key_exists($lid, $GmaLines)) {
          $GmaLines[$lid] = new GmaLine($lid);
        }
        $line = &$GmaLines[$lid];
        if ($opts['color']) $line->color = $opts['color'];
        if ($opts['thickness']) $line->color = $opts['thickness'];
        if ($opts['opacity']) $line->opacity = $opts['opacity'];

        $line->addPoint($opts['lat'], $opts['lon']);
        $ret .= ($GmaDebug) ?  "<pre>".Keep(print_r($line,1)) : '';
        break;

    // Gma Type: Map "(:gma-map [options]:)"
    case ('map'):
        global $GmaMap;
        global $MarkupFrame;
        // Create The Map Object and set attributes.
        $GmaMap = ($opts['id']) ? new GmaMap($opts['id']) : new GmaMap();
        if ($opts['ctrl']) $controls = (array) explode(',', $opts['ctrl']);
        $GmaMap->setControl($controls);
        if ($opts['view']) $GmaMap->view = $opts['view'];
        if ($opts['height']) $GmaMap->height = $opts['height'];
        if ($opts['width']) $GmaMap->width = $opts['width'];

        // Trigger the end-of-markup addition of the Javascript,
        // and give the map's target.
        if (!$MarkupFrame[0]['posteval']['mymarkup'])
          $MarkupFrame[0]['posteval']['mymarkup'] = 'gmaCleanup();';
        $ret = Keep("<div id='{$GmaMap->id}'></div>");
        $ret .= ($GmaDebug) ?  "<pre>".Keep(print_r($GmaMap,1)) : '';
        break;

    // Gma Type: Point "(:gma-point lat lon [options]:)"
    case ('point'):
        global $GmaPoints;
        // Build the Point object.
        $point = new GmaPoint($opts['lat'], $opts['lon']);
        $point->directions = gmaDefaults($opts['daddr'], $GmapDaddr);
        $point->icon = gmaDefaults($opts['icon'], '');
        $point->message .= gmaDefaults($opts['text'], '');
        if ($opts['link']) $ret .= $point->anchor($opts['link']);
        // Add the point to the Collection.
        array_push($GmaPoints, $point);
        $ret .= ($GmaDebug) ?  "<pre>".Keep(print_r($point,1)) : '';
        break;
    // Gma Type: (Invalid)
    default:
        $ret = "GMA Error: Type Unknown ($type)";
  }
  gmaSpoofLocale();
  return $ret;
}
function gmaSpoofLocale() {
    // (null) gmaSpoofLocale(null);
    //
    // Provided by HelgeLarsen. This allows for non-US locales to behave as
    // normal. I moved it into a function for smoother toggle, but this may be
    // OBE.
    global $GmaHoldLocale, $GmaLocaleToggle;
    if ($GmaLocaleToggle) {
      setlocale(LC_NUMERIC,$GmapHoldLocale);
    }
    else {
      $GmapHoldLocale = setlocale(LC_NUMERIC,'0');
      setlocale(LC_NUMERIC,'en_US');
    }
    $GmaLocaleToggle = ($GmaLocalToggle) ? 0 : 1;
}
function gmaZIP2LatLon($z) {
  // (array) gmaZIP2LatLon(int);
  //
  // This function allows a site to use ZIP codes for lat-lon, when configured.
  // Specifically, the function checks for the ZIP file and then greps the
  // answer.
  $zip_file = GMAPATH.'/library/gma_zip.csv';
  if (file_exists($zip_file) && $zip_file != '') {
    exec("grep '^$z' $zip_file", $r);
    $x = split(',', $r[0]);
    list($lat, $lon) = array($x[1], $x[2]);
    $lon = preg_replace('/^(.)0/', "$1", $lon);
    return array($lat, $lon);
  }
  return array(null,null); # No zip file.
}
function GmaDoIEFix() {
  global $GmapIEFix;
  if (get_cfg_var('browscap')) {
   $browser=get_browser(); //If available, use PHP native function
  }
  else {
   require_once(GMAPATH."library/php-local-browscap.php");
   $browser=get_browser_local(null, false, GMAPATH."/library/browscap.ini");
  }
  $GmapIEFix = (preg_match('/IE/', $browser->browser))
    ? " xmlns='http://www.w3.org/1999/xhtml' xmlns:v='urn:schemas-microsoft-com:vml'"
    : '';
}

